/* -*-mode:C;c-basic-offset:8-*-
 *      Copyright (C) 1994, 1995, 1996  Claus-Justus Heine
 *
 * This file implements a "generic" interface between the zftape-driver and a
 * compression-algorithm. The compression-algorithm currently used is a
 * LZ77. I use the implementation lzrw3 by Ross N. Williams
 * (Renaissance Software). The compression program itself is in the file
 * lzrw3.c and compress.h.
 * To adopt another compression algorithm the functions 
 * zftape_compress_info(), zftape_compress() and zftape_uncompress()
 * must be changed appropriately. See below.
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 $Author: claus $
 *
 $Revision: 1.20 $
 $Date: 1996/07/22 15:58:55 $
 $State: Exp $
 *
 */
static char RCSid[] =
"$Id: qic80-compress.c,v 1.20 1996/07/22 15:58:55 claus Exp $";


#include "ftape.h"
#include <asm/segment.h>
#include <linux/string.h>
#include <linux/errno.h>

#include "ftape-rw.h"
#include "ftape-ctl.h"
#include "ftape-read.h"
#include "ftape-write.h"
#include "qic80-compress.h"
#include "compress/compress.h"

/*
 *   global variables
 */

int ftape_total_data_amount        = 0;
int ftape_data_pos                 = 0;
int ftape_uncmpr_pos               = 0;
int ftape_seg_data_pos             = 0;
int zftape_use_compression          = 0;
int zftape_compression_map_changed  = 0;
int zftape_compression_map_location = 1;
int zftape_compression_map_length   = 0;
int ftape_first_user_segment       = -1;
DEFINE_BUFFER(byte, zftape_compression_wrk_mem, CMPR_WRK_MEM_SIZE );
DEFINE_BUFFER(byte, zftape_compression_buffer,  CMPR_BUFFER_SIZE );
/*
 *  I handle the allocation of this buffer as a special case, because it's
 *  size varies depending on the tape length inserted. In case that
 *  DYN_ALLOC is not defined, this buffer simply gets allocated statically.
 */
DEFINE_BUFFER(CompressionMapEntry, zftape_compression_map, COMPRESSION_MAP_SIZE/sizeof(CompressionMapEntry));
/*
 *   local variables
 */

static int compressed_size = 0; /* how much data is left in the 
                                 * zftape_compression_buffer
                                 */
static int compos          = 0; /* position inside zftape_compression_buffer */

/*
 * compressed segment:
 * 
 * 2 bytes: offset of compression segment structure
 *          29k > offset >= 29k-18: data from previous segment ens in this
 *                                  segment and no compressed block starts
 *                                  in this segment
 *                     offset == 0: data from previous segment occupies entire
 *                                  segment and continues in next segment
 * n bytes: remainder from previous segment
 * 6 bytes: compression segment structure:
 *          4 bytes: files set byte offset
 *          2 bytes: byte count N (amount of data following)
 *                   bit 15 is set if data is compressed, bit 15 is not
 *                   set if data is uncompressed
 * N bytes: data (as much as specified in the byte count)
 * 2 bytes: byte count N_1 of next cluster
 * N_1 bytes: data of next cluset
 * 2 bytes: byte count N_2 of next cluster
 * N_2 bytes: ...
 * 
 */

typedef struct {
	__u16 count;     /* amount of data as stored on tape */
	unsigned offset; /* offset in buffer */
} qic80_cluster;

typedef struct qic80_CompressionSegment
{
	__u16 cseg_first_block; /* location of header information in this 
				 * segment, i.e. of the cseg_offset
				 */
	unsigned cseg_rcount;   /* amount of data remaining from prev. seg */
	__u32 cseg_foffs;       /* file set byte offset, same as in 
				 * compression map segment
				 */
	qic80_cluster cseg_cluster;
} qic80_CSeg;

static qic80_CSeg cseg; /* static data. Must be kept uptodate by 
			 * read, write and seek functions
			 */

/*
 *  dispatch compression segment info, return error code
 */
static int
get_cseg ( byte *buff, int segment_size, qic80_CSeg *cseg, 
	   unsigned segment, int check )
{
	TRACE_FUN( 6, "get_cseg" );

	cseg->cseg_first_block = GET2( buff, 0 );
	if ( cseg->cseg_first_block == 0 ) {
		/*cseg->cseg_cluster.offset = segment_size;*/
		cseg->cseg_cluster.count = 0;
		cseg->cseg_rcount = segment_size - 2;
	} else if ( (segment_size - cseg->cseg_first_block) <= 18 ) {
	        cseg->cseg_rcount = cseg->cseg_first_block - 2;
		if ( check && cseg->cseg_rcount != compressed_size ) {
			TRACEx2( 1,"Compressed block should end "
				 "before byte %d, stored is %d.", 
				 compressed_size + 2, 
				 cseg->cseg_rcount + 2 );
			TRACE_EXIT;
			return -EIO;
		}
		cseg->cseg_cluster.count = 0;
		/*cseg->cseg_cluster.offset = segment_size;*/
	} else {
	        cseg->cseg_rcount = cseg->cseg_first_block - 2;
		if ( check && cseg->cseg_rcount != compressed_size ) {
			TRACEx2( 1, "First compressed block should "
				 "start at byte %d, stored is %d.", 
				 compressed_size + 8, 
				 cseg->cseg_first_block + 6 );
			TRACE_EXIT;
			return -EIO;
		}
		/*
		 * now get the offset the first block should have in 
		 * the uncompressed data stream. This is just a 
		 * consistency check, the information is already stored
		 * in the compression map.
		 */
		cseg->cseg_foffs  = GET4( buff, cseg->cseg_first_block );
		if ( cseg->cseg_foffs 
		     != 
		     zftape_compression_map[ segment ] ) {
			TRACEx3( 1, "compression_map (%d) != uncmpr. "
				 "pos (%d) stored in segment %d",
				 (int)zftape_compression_map[ segment ], 
				 (int)cseg->cseg_foffs, segment );
			TRACE_EXIT;
			return -EIO;
		}
		cseg->cseg_cluster.count = 0;
		cseg->cseg_cluster.offset = cseg->cseg_first_block + 4; 
	}
	TRACEi( 5,"first block                 : ",cseg->cseg_first_block );
	TRACEi( 5,"remainder from prev. segment: ",cseg->cseg_rcount);
	TRACEi( 5,"file set offset             : ",(int)cseg->cseg_foffs);
	TRACEi( 5,"start of first data block   : ",
		cseg->cseg_cluster.offset);
	TRACE_EXIT;
	return 0;
}

static inline void
get_next_cluster( byte *dbl_buff, int segment_size, qic80_cluster *cluster )
{
	TRACE_FUN( 6, "get_next_cluster");
	cluster->offset += (cluster->count&~0x8000);
	if ( (segment_size - cluster->offset) > 18 ) {
		cluster->count = GET2( dbl_buff, cluster->offset );
		cluster->offset += 2;
	} else {
		cluster->count = 0;
	}
	TRACEi( 5," offset: ", cluster->offset );
	TRACEi( 5,"  count: ", cluster->count  );
	TRACE_EXIT;
}

/*
 *  this function is needed for ftape_reset_position in ftape-io.c
 */

void
zftape_reset_cmpr_locals( void )
{
	compressed_size = 
	compos          = 0;
}


/******************************************************************************
 *
 *  The following three functions "ftape_compress_info()", "ftape_compress()"
 *  and "ftape_uncompress()" are the interface to the actual compression
 *  algorithm (i.e. they are calling the "compress()" function from the 
 *  lzrw3 package for now). These routines could quite easily be changed to 
 *  adopt another compression algorithm instead of lzrw3, which currently is
 *  used.
 *
 *****************************************************************************/

/*
 *  the purpose of this function is to return the amount of memory that
 *  is required by the compression-algorithm, because this memory
 *  is allocated dynamicaly (vmalloc) if DYN_ALLOC is defined. If 
 *  DYN_ALLOC is not defined this function must return a value that
 *  is smaller than or equal to CMPR_WRK_MEM_SIZE, defined in 
 *  "ftape-compress.h".
 *  If it returns a value greater than CMPR_WRK_MEM_SIZE, the driver failes to 
 *  initialize.
 */
#if 0
int
zftape_compress_info( void )
{        
	TRACE_FUN( 5, "ftape_compress_info");
	struct compress_identity *id;
	
	compress(COMPRESS_ACTION_IDENTITY, NULL, NULL, 0, NULL, 
		 (unsigned long *)&id);
	if ( id ) {
		printk("Id              : %08x\n"
		       "Required memory : %d bytes\n"
		       "Name            : %s\n"
		       "Version         : %s\n"
		       "Date            : %s\n"
		       "Copyright       : %s\n"
		       "Author          : %s\n"
		       "Affiliation     : %s\n"
		       "Vendor          : %s\n\n",
		       (unsigned int)id->id, (int)id->memory, id->name, 
		       id->version, id->date, id->copyright, id->author, 
		       id->affiliation, id->vendor);
	}        
	TRACE_EXIT;
	return id->memory;
}   
#endif
/*
 *  called by zftape_compress_write() to perform the compression. Must return
 *  the size of the compressed data.
 *
 *  NOTE: The size of the compressed data should not exceed the size of
 *        the uncompressed data. Most compression Algorithms have means to
 *        store data unchanged if the "compressed" data amount would exceed
 *        the original one. Mostly this is done by storing some flag-bytes
 *        in front of the compressed data to indicate if it is compressed or
 *        not. Thus the worst compression result length is the original length
 *        plus those flag-bytes. 
 *
 *        We don't want that, as the QIC-80 standard provides a means of 
 *        marking uncomrpessed blocks by simply setting bit 15 of the 
 *        compressed block's length. Thus a compessed block can have at most a 
 *        length of 2^15-1 bytes. The QIC-80 standard restricts the 
 *        block-length even further, allowing only 29k - 6 bytes.
 *
 *        Currently, the maximum blocksize used by zftape is 28k.
 *
 *        In short: don't exceed the length of the input-package, set bit 15 
 *        of the compressed size to 1 if you have copied data instead of 
 *        compressing it.
 *
 */
static unsigned short
zftape_compress( byte *in_buffer, int in_size, byte *out_buffer )
{ 
	TRACE_FUN( 5, "zftape_compress");
	unsigned long compressed_size;
	
	compress( COMPRESS_ACTION_COMPRESS,
		  zftape_compression_wrk_mem, in_buffer, in_size, out_buffer, 
		  &compressed_size);
	
#ifndef NO_TRACE_AT_ALL
	history.wr_uncompressed += in_size;
	history.wr_compressed += compressed_size & ~0x8000;
#endif
	
	TRACEx2( 7, "before compression: %d bytes, after compresison: %ld", 
		 in_size,        compressed_size & ~0x8000 );
	TRACE_EXIT;
	return (unsigned short)compressed_size;
}

/*
 * called by zftape_compress_read() to decompress the data. Must return the 
 * size of the decompressed data for sanity checks (compared with 
 * ftape_block_size)
 *
 *  NOTE: Read the note for zftape_compress() above!
 *        If bit 15 of the parameter in_size is set, then the data in in_buffer
 *        isn't compressed, which must be handled by the un-compression 
 *        algorithm. (I changed lzrw3 to handle this.)
 */
static unsigned short
zftape_uncompress( byte *in_buffer, unsigned short in_size, byte *out_buffer )
{ 
	TRACE_FUN( 5, "zftape_uncompress");
	unsigned long uncompressed_size;
	
	compress( COMPRESS_ACTION_DECOMPRESS,
		  zftape_compression_wrk_mem, in_buffer, (int)in_size, 
		  out_buffer, &uncompressed_size);
	
#ifndef NO_TRACE_AT_ALL
	history.rd_compressed += in_size & ~0x8000;
	history.rd_uncompressed += uncompressed_size;
#endif       
	
	TRACEx2( 7,"before decompression: %d bytes, after decompression: %ld", 
		 in_size & ~0x8000, uncompressed_size );
	TRACE_EXIT;
	return (unsigned short)uncompressed_size;
}


static int 
start_new_cseg( int *buf_pos_write, int this_segs_size )
{
	TRACE_FUN( 6, "start_new_cseg");
	int copy_count;
	int size_left;

	size_left = this_segs_size - 2 - compressed_size;
	TRACEi( 7, "segment size   : ", this_segs_size );
	TRACEi( 7, "compressed_size: ", compressed_size );
	TRACEi( 7, "size_left      : ", size_left );
	if ( size_left > 18 ) {
		copy_count = compressed_size;
		PUT2( deblock_buffer, 0, 2 + compressed_size );
		PUT4( deblock_buffer, 2 + compressed_size, 
		      ftape_uncmpr_pos + ftape_block_size);
		zftape_compression_map[ ftape_seg_pos ] = 
			ftape_uncmpr_pos + ftape_block_size;
		*buf_pos_write = 2 + compressed_size + 4;
	} else if ( size_left >= 0 ) {
		/* erase unused part of segment.
		 */
		memset( deblock_buffer + 2 + compressed_size, 
			'\0',  size_left );
		copy_count = compressed_size;
		PUT2( deblock_buffer, 0, 2 + compressed_size );  
		zftape_compression_map[ ftape_seg_pos ] =  0;
		*buf_pos_write = this_segs_size;
	} else {
		copy_count = this_segs_size - 2;
		if ( copy_count < 0 ) {
			copy_count = 0;
		}
		/*  indicate that we need entire segment and more space
		 */
		PUT2( deblock_buffer, 0, 0 ); 
		*buf_pos_write = this_segs_size;
		zftape_compression_map[ ftape_seg_pos ] =  0;
	}
	memcpy( deblock_buffer + 2, zftape_compression_buffer + compos,
		copy_count );
	TRACE_EXIT;
	return copy_count;
}


/*
 * return-value: the number of bytes removed from the user-buffer
 *
 * out: int *write_cnt: how much actually has been moved to the deblock_buffer
 *      int *req_len  : MUST NOT BE CHANGED, except at EOT, in which case it
 *                      MUST BE SET TO 0.
 *
 * in : char *buff          : the user buffer
 *      int buf_pos_write   : copy of buf_pos_wr from "_ftape_write()"
 *      int this_segs_size  : the size in bytes of the actual segment
 *      char *deblock_buffer: the deblock_buffer
 *
 */                         

int
zftape_compress_write ( int *write_cnt, int *req_len, const char *buff, 
                       int buf_pos_write, int this_segs_size, 
                       char *deblock_buffer)
{
	TRACE_FUN( 5, "zftape_compress_write");
	int req_len_left = *req_len;
	int remainder;
	int result;
	int len_left;
	int size;
	
	if (  ftape_data_pos >
	      (ftape_total_data_amount - (ftape_block_size + 8))
	      ) {
		TRACEx1( 7, "clipping req_len from %d to 0", *req_len );
		*req_len = req_len_left = 0;
	}    
	
	if ( buf_pos_write == 0 ) { /* fill a new segment */
		remainder = start_new_cseg( &buf_pos_write, this_segs_size );
		compressed_size -= remainder; /* amount left in the buffer */
		if ( compressed_size != 0 ) {
			compos  += remainder; 
			result   = 0;
		} else if ( compos  != 0 ) {
			req_len_left -= result = ftape_block_size;
			compos  = 0;
		} else {
			result = 0;
		}
		*write_cnt = buf_pos_write;
	} else {
		*write_cnt = result = 0;
	}
	
	len_left = this_segs_size - buf_pos_write;
	while ( (req_len_left > 0) && (len_left > 18) ) {
		/* now we have some size left for a new compressed block. 
		 * We know, that the compression buffer is empty (else there 
		 * wouldn't be any space left).
		 */
		memcpy_fromfs( scratch_buffer, buff+result, ftape_block_size);
		req_len_left -= ftape_block_size;
		TRACEx2( 7, "Compressing block %d, "
			 "data-position (uncompressed) %d bytes.",
			 ftape_uncmpr_pos/ftape_block_size,ftape_uncmpr_pos);
		compressed_size= zftape_compress( scratch_buffer,
						  ftape_block_size, 
						  zftape_compression_buffer );
		PUT2( deblock_buffer, buf_pos_write, compressed_size );
		compressed_size &= ~0x8000; /* strip non-compression flag */
		buf_pos_write += 2;
		*write_cnt    += 2;
		len_left -= 2;
		/* increment "result" iff we copied the entire compressed 
		 * block to the deblock_buffer
		 */
		if ( len_left >= compressed_size ) {
			size      = compressed_size;
			len_left -= compressed_size;
			compos = compressed_size = 0;
			result += ftape_block_size;
		} else {
			size = compos = len_left;
			compressed_size -= len_left;
			len_left = 0;
		}
		memcpy( deblock_buffer + buf_pos_write, 
			zftape_compression_buffer, size );
		*write_cnt += size;
		buf_pos_write   += size;
	}
	/*  increment global data offset in current volume
	 */
	if ( result > 0 ) {
		ftape_uncmpr_pos    += result;
		TRACEi( 7, "ftape_uncmr_pos", ftape_uncmpr_pos );
	}
	/*  erase the remainder of the segment if less than 18 bytes left
	 *  (18 bytes is due to the QIC-80 standard)
	 */
	if ( len_left <= 18 ) {
		memset( deblock_buffer + buf_pos_write, '\0', len_left );
		(*write_cnt) += len_left;
	}
	TRACEx2( 7, "compressed_size: %d, compos : %d", 
		 compressed_size, compos );
	TRACE_EXIT;
	return result;
}   


/*
 * out:
 *
 * int *read_cnt: the number of bytes we removed from the deblock_buffer
 *                (result)
 * int *to_do   : the remaining size of the read-request.
 *
 * in:
 *
 * char *buff   : buff is the address of the upper part of the user buffer,
 *                that hasn't been filled with data yet.
 * int buf_pos_read:  copy of from _ftape_read()
 * int buf_len_read:  copy of buf_len_rd from _ftape_read()
 * char *deblock_buffer: deblock_buffer
 * unsigned short block_size: the block size valid for this volume, may differ
 *                            from ftape_block_size.
 *
 * returns the amount of data actually copied to the user-buffer
 *
 * *to_do MUST NOT SHRINK except to indicate an EOF. In this case *to_do has to
 * be set to 0
 *
 */

int
zftape_compress_read ( int *read_cnt, int to_do, char *buff, 
                      int buf_pos_read, int segment_size, 
                      char *deblock_buffer, unsigned short block_size )
{          
	TRACE_FUN( 5, "ftape_compress_read");
	int uncompressed_size;         
	int result = 0;
	static int stored_size;

	if ( buf_pos_read == 0 ) {
		/*
		 *  new segment just read
		 */
		TRACEx1( 7, "carry from previous block: %d", compressed_size );
		if ( (result = get_cseg ( deblock_buffer, segment_size, 
					  &cseg, ftape_seg_pos, 1 ) ) < 0 ) {
			*read_cnt = 0;
			TRACE_EXIT;
			return result;
		}
		memcpy( zftape_compression_buffer + compos,
			deblock_buffer + 2, cseg.cseg_rcount );
		*read_cnt         = cseg.cseg_rcount + 2 + 4;
		compos           += cseg.cseg_rcount;
		compressed_size  -= cseg.cseg_rcount;
	} else {
		*read_cnt = 0;
	}

	/*
	 *  loop and uncompress until user buffer full or deblock-buffer empty
	 */
	
	TRACEx3( 5, "compressed_size: %d, compos : %d, *read_cnt: %d",
		 compressed_size, compos, *read_cnt );
	while ( (compressed_size == 0) && (to_do > 0) ) {
		if ( compos  != 0 ) { 
			uncompressed_size = 
				zftape_uncompress( zftape_compression_buffer,
						   stored_size,
						   scratch_buffer );
			if ( uncompressed_size != block_size ) {
				TRACEx2( 1,"Uncompressed block (%d) "
					 "!= block-size (%d)",
					 uncompressed_size, 
					 block_size);
				*read_cnt = 0;
				TRACE_EXIT;
				return -EIO;
			}       
			memcpy_tofs( buff + result, scratch_buffer, 
				     uncompressed_size );
			to_do      -= uncompressed_size;
			result     += uncompressed_size;
			compos  = 0;
		}                                              
		if ( to_do > 0 ) {
			get_next_cluster( deblock_buffer, 
					  segment_size, 
					  &cseg.cseg_cluster );
			if ( cseg.cseg_cluster.count != 0 ) {
				stored_size     = cseg.cseg_cluster.count;
				compressed_size = 
					cseg.cseg_cluster.count & ~0x8000;
				compos = segment_size-cseg.cseg_cluster.offset;
				if ( compos > compressed_size ) {
					compos = compressed_size;
				}
				memcpy(zftape_compression_buffer , 
				       deblock_buffer+cseg.cseg_cluster.offset,
				       compos );
				TRACEi( 5, "just copied to cmpr-buffer: ", 
					compos);
				compressed_size -= compos ;
				*read_cnt += compos + 2;
			} else {
				to_do = 0;
			}
		}
		/*  Allow escape from this loop on signal !
		 */
		if (current->signal & _DONT_BLOCK) {
			TRACE( 2, "interrupted by non-blockable signal");
			TRACE_EXIT;
			return -EINTR;
		}
		TRACEx3( 5, "compressed_size: %d, compos : %d, *read_cnt: %d",
		 compressed_size, compos, *read_cnt );
	}
	if ( (segment_size - *read_cnt - buf_pos_read) <= 18 ) {
		*read_cnt = segment_size - buf_pos_read;
		TRACEi( 5, "expanding read cnt to: ", *read_cnt );
	}
	TRACEi( 5, "segment size: ", segment_size );
	TRACEi( 5, "read count  : ", *read_cnt );
	TRACEi( 5, "buf_pos_read: ", buf_pos_read );
	TRACEi( 5, "remaining   : ", segment_size - *read_cnt - buf_pos_read );
	TRACEx2( 7, "compressed_size: %d, compos : %d", 
		 compressed_size, compos );
	ftape_uncmpr_pos += result;
	TRACEi( 7, "ftape_uncmpr_pos: ",ftape_uncmpr_pos );
	TRACEi( 5, "returning: ", result );
	TRACE_EXIT;
	return result;
}                

/*
 *  seeks to the new data-position. Reads sometimes a segment.
 *  This function uses the compression map.
 *  
 *  start_seg and end_seg give the boundaries of the current volume
 *  block_size is the block_size of the current volume as stored in the
 *  volume label
 *
 */
int
zftape_compress_seek( int new_block_pos, 
                     unsigned start_seg, 
                     unsigned end_seg, 
                     unsigned block_size )
{ 
	TRACE_FUN( 5, "zftape_compress_seek");
	int result = 0;
	int new_pos;
	int desired_pos;
	int biggest;              
	int new_seg_pos;
	int this_segs_size;

	compressed_size = compos = 0;
	TRACEx1( 7, "desired new block positon: %d", new_block_pos );
	if ( new_block_pos <= 0 ) { /* handle it fast, nothing to do */
		ftape_seg_pos      = start_seg;
		ftape_seg_data_pos = ftape_uncmpr_pos   = 0;
	} else {
		desired_pos = new_block_pos * block_size;
		TRACEx1( 5, "desired new position: %d bytes", desired_pos);
		/*
		 *   get the biggest segment-id with compression map entry != 0
		 */
		for ( biggest = end_seg;
		      (biggest > start_seg) && 
		      (zftape_compression_map[ biggest ] == 0);
		      biggest -- );
		TRACEi( 5, "biggest:", biggest);
		/*
		 * now get the biggest segment id, that has an 
		 * compression-map entry != 0
		 * which is >= the desired position
		 */
		for ( new_seg_pos = start_seg;
		      new_seg_pos < biggest && 
		      desired_pos > zftape_compression_map[new_seg_pos];
		      new_seg_pos++ ) {
			TRACEx2( 5, "segment: %d, cmap: %d", new_seg_pos,
				 (int)zftape_compression_map[new_seg_pos] );
		}
		/* 
		 * We may have skipped some segments with zero compression 
		 * map entry.
		 * Now go back until we find a segment with an compression map
		 * entry that is <= new_umcmpr_pos.
		 */
		while ( (new_seg_pos > start_seg) && 
			(zftape_compression_map[new_seg_pos] > desired_pos )) {
			TRACEx2( 5, "segment: %d, cmap: %d", new_seg_pos,
				 (int)zftape_compression_map[new_seg_pos] );
			new_seg_pos --;
		} 
		TRACEx2( 5, "last entry in compression_map: %d/%d",
			 biggest, (int)zftape_compression_map[biggest]);
		TRACEx2( 5, "New segment:                   %d/%d",
			 new_seg_pos, 
			 (int)zftape_compression_map[new_seg_pos]);
		TRACEi( 5, "desired new position:", desired_pos);
		/*
		 * now biggest is the biggest entry != 0 and new_seg_pos is 
		 * the segment, that desired_pos belongs to
		 */     
		this_segs_size = ftape_get_seg_size( new_seg_pos );
		/*
		 * read segment if necessary
		 * NOTE: ftape_deblock_segment stores the segment id of the 
		 * segment that is held in the deblock_buffer, if any. Else 
		 * ftape_deblock_segment is set to -1
		 */
		if ( new_seg_pos != ftape_deblock_segment ) {
			result = read_segment( new_seg_pos, deblock_buffer, 0);
			if ( result != this_segs_size ) {
				TRACEx2( 1,"Can't seek to byte %d, "
					 "segment %d is unreachable.",
					 desired_pos, new_seg_pos );
				ftape_deblock_segment = -1;
				TRACE_EXIT;
				return -EIO;
			}
			ftape_deblock_segment = new_seg_pos;
		}
		/* get the compression structures
		 */
		if ( (result = get_cseg ( deblock_buffer, this_segs_size, 
					  &cseg, new_seg_pos, 0) ) < 0 ) {
			TRACEi( 1, "Error in get_cseg: ", result );
			TRACE_EXIT;
			return result;
		}
		/* loop while current position is valid, i.e. inside current 
		 * segment and position stored in segment < desired position
		 */
		new_pos = cseg.cseg_foffs;
		TRACEi( 5, "offset : ", cseg.cseg_cluster.offset );
		TRACEi( 5, "segsize: ", this_segs_size );
		TRACEi( 5, "desired: ", desired_pos );
		TRACEi( 5, "actual : ", new_pos ); 
		while ( (cseg.cseg_cluster.offset < this_segs_size) && 
			(new_pos < desired_pos) ) {
			get_next_cluster( deblock_buffer, this_segs_size, 
					  &cseg.cseg_cluster );
			if ( cseg.cseg_cluster.count == 0 ) { 
				/* end of recorded media */
				desired_pos = new_pos;
			} else {
				new_pos += block_size;
			}
			TRACEi( 5, "new position:", new_pos );
		}
		if ( new_pos != desired_pos ) {
			result = -EIO; /* after EOF */
		} 
		ftape_seg_data_pos =   cseg.cseg_cluster.offset 
			             + cseg.cseg_cluster.count;
		ftape_seg_pos      = new_seg_pos;
		ftape_uncmpr_pos   = new_pos;
		TRACEi( 7, "ftape_uncmpr_pos:", ftape_uncmpr_pos );
		TRACEi( 7, "ftape_seg_pos:", ftape_seg_pos);
		TRACEi( 7, "ftape_seg_data_pos:", ftape_seg_data_pos);
	}
	ftape_data_pos = ftape_calc_data_pos(ftape_seg_pos,ftape_seg_data_pos);
	TRACEi( 7, "ftape_data_pos:", ftape_data_pos );
	TRACE_EXIT;
	return result;
}            


/*
 *  extract the compression map from the compression-map segments, starting
 *  at "location". Support for 1100 ft. tapes is added, but untested, as I 
 *  don't have one.
 * 
 */
int
zftape_extract_compression_map( byte *address, unsigned location )
{
	TRACE_FUN( 5, "extract_compression_map");
	int remaining;
	int result= 0;
	int offset;
	/*
	 *  initialize variables in case we can't extract the compression map.
	 *  first_data_segment holds the volume-table
	 */
	zftape_compression_map_length= 0;     
	ftape_first_user_segment = first_data_segment + 1;
	if( location > 0 ) {
		/*
		 *  location == 0 means no compression map.
		 */
		TRACEi( 4, "compression_map_location: ", location );
		/*
		 *  we read ahead in case the DCMS fills more than one segment
		 */
		result = read_segment( location, address, 1);
		if( result < 0 ) {
			TRACEi( 1, "Couldn't read compression map segment at", 
				location );
			TRACE_EXIT;
			return -EIO ;
		}
		if (memcmp(address, "DCMS", 4) != 0) {
			TRACEi( 1, "Wrong signature of "
				"compression map segment at", location );
			TRACE_EXIT;
			return -EIO;
		}
		remaining = zftape_compression_map_length = 
			segments_per_track * tracks_per_tape;
		remaining *= sizeof(CompressionMapEntry);
#if 0
		if ( zftape_compression_map_length 
		     != 
		     (segments_per_track * tracks_per_tape) ) {
			zftape_compression_map_length = 0;
			TRACE( 1, "compression map stored on tape "
			       "has wrong size");
			TRACE_EXIT;
			return -EIO;
		}
#endif
		if ( zftape_allocate_compression_map() < 0 ) {
			return -ENOMEM;
		}
#if 0 && !defined(NO_TRACE_AT_ALL)
		/*
		 *  just a little more noise
		 */
		switch( zftape_compression_map_length ) {
		case MAP_205_FEET  : {
			TRACEx1( 4, "Found compression map segment "
				 "at %d for 205 feet tape", location );
			break;
		}
		case MAP_307_5_FEET: {
			TRACEx1( 4, "Found compression map segment "
				 "at %d for 307,5 feet tape",location);
			break;
		}
		case MAP_770_FEET: { 
			TRACEx1( 4, "Found compression map segment "
				 "at %d for 770 feet tape",location);
			break;
		}
		case MAP_1100_FEET : {
			TRACEx1( 4, "Found compression map segment "
				 "at %d for 1100 feet tape.",location);
			break;
		}
		default   : {
			TRACEx1( 4, "Got compression map with "
				 "%d entries, tape is supposed to "
				 "have that many segments", 
				 zftape_compression_map_length );
			break;
		}
		}
#endif
		/*
		 *  copy data from buffer to compression map
		 *  read in segments until we got the entire compression map
		 */
		do {
			if ( result > remaining ) {
				result = remaining;
			}
			offset =   zftape_compression_map_length 
				 - remaining / sizeof(CompressionMapEntry);
			memcpy( zftape_compression_map + offset, address, 
				result );
			remaining -= result;
			if ( remaining > 0 ) {
				result = read_segment( ++location, address, 1);
				if( result < 0 ) {
					TRACEi( 1, "Couldn't read compression "
						"map segment at", location );
					TRACE_EXIT;
					return -EIO ;
				}
			}
		} while ( remaining > 0 );
		/*
		 *  first segment after compression map segments
		 */
		ftape_first_user_segment = location + 1;
	}
	TRACE_EXIT;
	return 0;
}




/*
 *  write one or more compression map segments starting at "segment"
 *  using buffer. Adjusts ftape_first_user_segment
 *
 */

int
zftape_update_compression_segments( unsigned segment, byte* buffer )
{
	TRACE_FUN( 5, "update_compression_segments");
	int result = 0;
	int remaining;
	int written;
	int rel_seg = 0;

	if(segment == 0 || zftape_compression_map == NULL ||
	   segment == header_segment_1 || segment == header_segment_2 ) {
		TRACE_EXIT;
		return -EINVAL;
	} 
	TRACEx1( 4, "Creating or updating compression-map-segment at %d", 
		 segment);
	memset( (byte *)zftape_compression_map + 6, '\0', 6 );
	memcpy( zftape_compression_map, "DCMS", 4 );
	PUT2( (byte *)zftape_compression_map, 4, 
	      sizeof(CompressionMapEntry)*zftape_compression_map_length );
	/*
	 * zftape_compression_map_length holds the number of segments on tape,
	 * that is tracks_per_tape * segments_per_track
	 */
	remaining = zftape_compression_map_length * sizeof(CompressionMapEntry)
		    - SIZEOF(deblock_buffer);
	if ( remaining >= 0 ) {
		memcpy( buffer, zftape_compression_map,SIZEOF(deblock_buffer));
	} else {
		memcpy( buffer, zftape_compression_map,
			sizeof(CompressionMapEntry)
			*zftape_compression_map_length );
		(void)memset(buffer+sizeof(CompressionMapEntry)
			     *zftape_compression_map_length,
			     '0', -remaining);
		remaining = 0;
	}
	/*
	 *  now write one or more compression map segments
	 *
	 */
	while ( (result = _write_segment( segment, buffer, 0 )) == -EAGAIN);
	written = result;
	while ( result >= 0 && written < (zftape_compression_map_length*4) ) {
		if ( result < SIZEOF(deblock_buffer) ) {
			memmove( buffer, buffer + result, 
				 SIZEOF(deblock_buffer) - result );
		}
		if ( remaining < result ) {
			memcpy( 
			       buffer + SIZEOF(deblock_buffer) - result,
			       zftape_compression_map 
			       + zftape_compression_map_length 
			       - (remaining/sizeof(CompressionMapEntry)), 
			       remaining );
			memset( buffer + SIZEOF(deblock_buffer) 
				- result + remaining,
				'\0',
				result - remaining );
			remaining = 0;
		} else {
			memcpy( 
			       buffer + SIZEOF(deblock_buffer) - result,
			       zftape_compression_map 
			       + zftape_compression_map_length 
			       - (remaining/sizeof(CompressionMapEntry)),
			       result );
			remaining -= result;
		}
		result = _write_segment( ++rel_seg + segment, buffer, 0 );
		written += result;
	}
	/*
	 *   this is the first segment usable for data:
	 */
	ftape_first_user_segment = segment + rel_seg + 1;
	TRACE_EXIT;
	return result;
}

int
zftape_verify_compression_map( unsigned segment, byte* buffer )
{
	TRACE_FUN( 5, "zftape_verify_compression_map");
	int remaining;
	int read_in = 0, to_read;

	/*
	 * zftape_compression_map_length holds the number of segments on tape,
	 * that is tracks_per_tape * segments_per_track
	 */
	remaining = zftape_compression_map_length*sizeof(CompressionMapEntry);
	do {
		to_read = remaining;
		remaining -= ftape_get_seg_size( segment );
		if ( remaining >= 0 ) {
			to_read = ftape_get_seg_size( segment );
		} 
		read_in = zftape_verify_read_segment( segment, buffer, 
				       (byte *)zftape_compression_map+read_in,
						      to_read, 0 );
		if ( read_in < to_read ) {
			TRACE_EXIT;
			return -EIO;
		}
		segment ++;
	} while ( remaining > 0 );
	TRACE_EXIT;
	return 0;
}
			

